

#include <string>
#include "Gar3DClass.hpp"
#include "Graphics.hpp"

namespace gar3d {

	Gar3D::Gar3D() : Gar3D(1280, 720) {}

	Gar3D::Gar3D(int width, int height) {
		int status = SDL_Init(SDL_INIT_VIDEO);
		if (status < 0)
			N_Graphics::die_with_sdl_error("Failed to initialize SDL");
		
		window = N_Graphics::createWindow("Gar3D", width, height);
		
		_input.userQuit = false;
		_orbitcam = OrbitCamera(10, 0.3, 0.1, Perspective(30, 16 / 9., 1, 20));
		_pancam = PanCamera(vec3(0, 4, 8), vec3(-1, 0, 0), vec3(0, 0, -1), glm::normalize(vec3(0, -1, -2)), Perspective(30, 16 / 9., 1, 20));
		_freecam = FreeCamera();
		_camera = &_orbitcam;
		_pancam.minbounds = vec3(-10, -10, -10);
		_pancam.maxbounds = vec3( 10,  10,  10);
		_lastFrame = Timepoint();
		_brightness = 0.8;
	}

	Gar3D::~Gar3D() {
		SDL_DestroyWindow(window);
		SDL_Quit();
	}

	bool Gar3D::shouldQuit() {
		return _input.userQuit;
	}

	void Gar3D::addLight(GLenum light, glm::vec4 position, vec3 color) {
		glEnable(light);
		glLightfv(light, GL_POSITION, &position[0]);
		glLightfv(light, GL_DIFFUSE, &color[0]);
	}

	void Gar3D::renderFrame() {
		_input.updateState();

		// Light gray background
		Color c = _backgroundColor;
		glClearColor(c.red(), c.green(), c.blue(), 1);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glEnable(GL_LIGHTING);
		addLight(GL_LIGHT0, glm::vec4(-100, 100, -100, 1), vec3(_brightness, _brightness, _brightness) / 2);
		addLight(GL_LIGHT1, glm::vec4(100, -100, -100, 1 ), vec3(_brightness, _brightness, _brightness) / 4);
		addLight(GL_LIGHT2, glm::vec4(100, 100, 100, 1), vec3(_brightness, _brightness, _brightness) / 3);
		addLight(GL_LIGHT3, glm::vec4(-100, -100, 100, 1), vec3(_brightness, _brightness, _brightness) / 3);

		_camera->apply(_input);

		/*
		// light source at camera
		glEnable(GL_LIGHTING);
		vec3 pos = _camera->getPosition();
		addLight(GL_LIGHT0, glm::vec4(pos.x, pos.y, pos.z, 1), lightlevel);
		/*/

		//  draw all objects
		for (int i = 0; i < _objbuffer.size(); i++) {
			renderPrimitive(_objbuffer[i], _oribuffer[i], _locbuffer[i], _sizbuffer[i], _colbuffer[i]);
		}
		glBegin(GL_LINES);
		for (int i = 0; i < _linebuffer.size(); i++) {
			glColor3f(_linecolbuffer[i].red(), _linecolbuffer[i].green(), _linecolbuffer[i].blue());
			glVertex3f(_linebuffer[i].x, _linebuffer[i].y, _linebuffer[i].z);
		}
		glEnd();
		_objbuffer.clear();
		_oribuffer.clear();
		_locbuffer.clear();
		_sizbuffer.clear();
		_colbuffer.clear();
		_linebuffer.clear();
		_linecolbuffer.clear();
		SDL_GL_SwapWindow(window);

		Timepoint now;
		_deltatime = (now - _lastFrame).Seconds();
		_lastFrame = now;
	}

	/*void Gar3D::stall(double duration) {
		Timepoint begin;
		Timepoint end = begin + Duration(duration, TimeUnit::Second);
		while (true) {
			Timepoint now;
			if (now >= end) {
				break;
			}
			_input.pollEvents();
		}
	}//*/

	void Gar3D::drawLine(double x0, double y0, double z0, double x1, double y1, double z1) {
		_linebuffer.push_back(vec3(x0, y0, z0));
		_linebuffer.push_back(vec3(x1, y1, z1));
		_linecolbuffer.push_back(_objColor);
		_linecolbuffer.push_back(_objColor);
	}

	void Gar3D::drawPrimitive(primitive prim, orientation orien, double posx, double posy, double posz, double size) {
		drawPrimitive(prim, orien, posx, posy, posz, size, size, size);
	}

	void Gar3D::drawPrimitive(primitive prim, orientation orien, double posx, double posy, double posz, double sizex, double sizey, double sizez) {
		_objbuffer.push_back(prim);
		_oribuffer.push_back(orien);
		_locbuffer.push_back(vec3(posx, posy, posz));
		_sizbuffer.push_back(vec3(sizex, sizey, sizez));
		_colbuffer.push_back(_objColor);
	}

	void Gar3D::renderPrimitive(primitive prim, orientation orien, vec3 location, vec3 size, Color color) {
		glColor3f(color.red(), color.green(), color.blue());
		glPushMatrix();
		glTranslatef(location.x, location.y, location.z);
		glScalef(size.x, size.y, size.z);
		switch (orien) {
		default:
		case positive_z:
			break;
		case positive_y:
			glRotatef(-90, 1, 0, 0);
			break;
		case positive_x:
			glRotatef(-90, 0, 1, 0);
			break;
		case negative_z:
			glRotatef(180, 0, 1, 0);
			break;
		case negative_y:
			glRotatef(90, 1, 0, 0);
			break;
		case negative_x:
			glRotatef(90, 0, 1, 0);
			break;
		}

		switch (prim) {
		case sphere:
			glScalef(0.5, 0.5, 0.5);
			renderSphere();
			break;
		case box:
			renderCube();
			break;
		case cylinder:
			renderCylinder();
			break;
		case cone:
			renderPyramid(30);
			break;
		case pyramid_3:
			renderPyramid(3);
			break;
		case pyramid_4:
			renderPyramid(4);
			break;
		case pyramid_6:
			renderPyramid(6);
			break;
		case plane:
			renderPlane();
			break;
		case arrow:
			renderArrow(0.05);
			break;
		case line:
			renderLine();
			break;
		/*case circle:
			renderCircle(30);
			break;//*/
		}

		glPopMatrix();
	}

	void Gar3D::renderSphere() {
		GLUquadric* quadric = gluNewQuadric();
		gluSphere(quadric, 1, 30, 30);
		gluDeleteQuadric(quadric);
	}

	void Gar3D::renderCube() {
		const vec3 cubeverts[8] = {
			vec3(0,0,0),
			vec3(1,0,0),
			vec3(0,1,0),
			vec3(1,1,0),
			vec3(0,0,1),
			vec3(1,0,1),
			vec3(0,1,1),
			vec3(1,1,1)
		};
		const int faces[6][4] = {
			{ 0,2,3,1 },
			{ 0,4,6,2 },
			{ 0,1,5,4 },
			{ 4,5,7,6 },
			{ 1,3,7,5 },
			{ 2,6,7,3 }
		};
		glTranslatef(-.5, -.5, -.5);
		glBegin(GL_QUADS);
		for (int f = 0; f < 6; f++) {
			for (int i = 0; i < 4; i++) {
				vec3 v = cubeverts[faces[f][i]];
				glVertex3f(v.x, v.y, v.z);
			}
		}
		glEnd();
	}

	void Gar3D::renderCylinder() {
		glTranslatef(0, 0, -0.5);
		GLUquadric* quadric = gluNewQuadric();
		gluCylinder(quadric, .5, .5, 1, 30, 1);
		gluDeleteQuadric(quadric);
	}

	void Gar3D::renderPyramid(int sides) {
		glTranslatef(0, 0, -0.5);
		GLUquadric* quadric = gluNewQuadric();
		gluCylinder(quadric, 0.5, 0, 1, sides, 1);
		gluDeleteQuadric(quadric);
	}

	void Gar3D::renderPlane() {
		glBegin(GL_QUADS);
		glVertex3f(-.5, -.5, 0);
		glVertex3f(-.5, .5, 0);
		glVertex3f(.5, .5, 0);
		glVertex3f(.5, -.5, 0);
		glEnd();
	}

	void Gar3D::renderArrow(float thick) {
		glPushMatrix();
		GLUquadric* quadric = gluNewQuadric();
		gluCylinder(quadric, thick, thick, 1 - 6 * thick, 30, 1);
		glTranslatef(0, 0, 1 - 6 * thick);
		gluCylinder(quadric, 2 * thick, 0, 6 * thick, 30, 1);
		gluDeleteQuadric(quadric);
		glPopMatrix();
	}

	void Gar3D::renderLine() {
		glBegin(GL_LINES);
		glVertex3f(0, 0, 0);
		glVertex3f(0, 0, 1);
		glEnd();
	}

	/*void Gar3D::renderCircle(int segments) {
		glBegin(GL_LINE_LOOP);
		for (int i = 0; i < segments; i++) {
			float theta = i / segments * 2 * M_PI;
			glVertex3f(cos(theta)/2, 0, sin(theta)/2);
		}
		glEnd();
	}//*/
	
	
	void Gar3D::getFreeCameraPosition(double& x, double& y, double& z) { 
		x = _freecam.position.x;
		y = _freecam.position.y;
		z = _freecam.position.z;
	}
	
	void Gar3D::setFreeCameraPosition(double x, double y, double z) { 
		_freecam.position.x = x;
		_freecam.position.y = y;
		_freecam.position.z = z;
	}
	
	void Gar3D::getFreeCameraTarget(double& x, double& y, double& z) {
		x = _freecam.target.x;
		y = _freecam.target.y;
		z = _freecam.target.z;
	}
	
	void Gar3D::setFreeCameraTarget(double x, double y, double z) { 
		_freecam.target.x = x;
		_freecam.target.y = y;
		_freecam.target.z = z;
	}
	
	void Gar3D::getPanCameraMinBounds(double& x, double& y, double& z) {
		x = _pancam.minbounds.x;
		y = _pancam.minbounds.y;
		z = _pancam.minbounds.z;
	}
	
	void Gar3D::setPanCameraMinBounds(double x, double y, double z) { 
		_pancam.minbounds.x = x;
		_pancam.minbounds.y = y;
		_pancam.minbounds.z = z;
	}
	
	void Gar3D::getPanCameraMaxBounds(double& x, double& y, double& z) { 
		x = _pancam.maxbounds.x;
		y = _pancam.maxbounds.y;
		z = _pancam.maxbounds.z;
	}
	
	void Gar3D::setPanCameraMaxBounds(double x, double y, double z) {
		_pancam.maxbounds.x = x;
		_pancam.maxbounds.y = y;
		_pancam.maxbounds.z = z;
	}
	
	void Gar3D::getActiveCameraPosition(double& x, double& y, double& z) {
		vec3 pos = _camera->getPosition();
		x = pos.x;
		y = pos.y;
		z = pos.z;
	}
	
	cameraMode Gar3D::CameraMode() {
		return(_camera == &_orbitcam) ? orbitcam : (_camera == &_pancam) ? pancam : freecam;
	}
	
	cameraMode Gar3D::CameraMode(cameraMode mode) {
		_camera = (mode == orbitcam) ? &_orbitcam : (mode == pancam) ? (Camera*)&_pancam : &_freecam;
		return mode;
	}

}  // namespace gar3d





















/*

void Gar3D::die_with_sdl_error(std::string message) {
	std::cout << message << ": " << SDL_GetError() << std::endl;
	exit(EXIT_FAILURE);
}

void Gar3D::die_if_opengl_error() {
	GLenum error = glGetError();
	if (error != GL_NO_ERROR) {
		std::cout << "OpenGL error: " << gluErrorString(error) << std::endl;
		exit(EXIT_FAILURE);
	}
}


VertexBuffer Gar3D::allocateVertexBuffer(int size) {
GLuint buffer;
glGenBuffers(1, &buffer);
glBindBuffer(GL_ARRAY_BUFFER, buffer);
glBufferData(GL_ARRAY_BUFFER, size, NULL, GL_STATIC_DRAW);
die_if_opengl_error();
return buffer;
}

ElementBuffer Gar3D::allocateElementBuffer(int size) {
GLuint buffer;
glGenBuffers(1, &buffer);
glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer);
glBufferData(GL_ELEMENT_ARRAY_BUFFER, size, NULL, GL_STATIC_DRAW);
die_if_opengl_error();
return buffer;
}

void Gar3D::copyVertexData(VertexBuffer buffer, void *data, int size) {
glBindBuffer(GL_ARRAY_BUFFER, buffer);
glBufferSubData(GL_ARRAY_BUFFER, 0, size, data);
die_if_opengl_error();
}

void Gar3D::copyElementData(ElementBuffer buffer, void *data, int size) {
glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer);
glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, size, data);
die_if_opengl_error();
}

void Gar3D::setVertexArray(VertexBuffer buffer) {
glEnableClientState(GL_VERTEX_ARRAY);
glBindBuffer(GL_ARRAY_BUFFER, buffer);
glVertexPointer(3, GL_FLOAT, 0, 0);
die_if_opengl_error();
}

void Gar3D::setColorArray(VertexBuffer buffer) {
glEnableClientState(GL_COLOR_ARRAY);
glBindBuffer(GL_ARRAY_BUFFER, buffer);
glColorPointer(3, GL_FLOAT, 0, 0);
die_if_opengl_error();
}

void Gar3D::setNormalArray(VertexBuffer buffer) {
glEnableClientState(GL_NORMAL_ARRAY);
glBindBuffer(GL_ARRAY_BUFFER, buffer);
glNormalPointer(GL_FLOAT, 0, 0);
die_if_opengl_error();
}

void Gar3D::setTexCoordArray(VertexBuffer buffer) {
glEnableClientState(GL_TEXTURE_COORD_ARRAY);
glBindBuffer(GL_ARRAY_BUFFER, buffer);
glTexCoordPointer(2, GL_FLOAT, 0, 0);
die_if_opengl_error();
}

void Gar3D::drawElements(GLenum mode, ElementBuffer buffer, int count) {
glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer);
glDrawElements(mode, count, GL_UNSIGNED_INT, 0);
die_if_opengl_error();
}

Texture Gar3D::loadTexture(std::string bmpFile) {
SDL_Surface *surface = SDL_LoadBMP(bmpFile.c_str());
if (surface == NULL)
die_with_sdl_error("Failed to load image");
SDL_LockSurface(surface);
GLuint texture;
glGenTextures(1, &texture);
glBindTexture(GL_TEXTURE_2D, texture);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, surface->w, surface->h, 0, GL_BGR,
GL_UNSIGNED_BYTE, surface->pixels);
glGenerateMipmap(GL_TEXTURE_2D);
SDL_FreeSurface(surface);
die_if_opengl_error();
return texture;
}

void Gar3D::setTexture(Texture texture) {
glEnable(GL_TEXTURE_2D);
glBindTexture(GL_TEXTURE_2D, texture);
die_if_opengl_error();
}

void Gar3D::unsetTexture() {
glDisable(GL_TEXTURE_2D);
}


//*/
